/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_GFX_PAINT_HPP
#define FGGL_GFX_PAINT_HPP

#include <string>
#include <vector>
#include <fggl/math/types.hpp>

namespace fggl::gfx {

	using RadianAngle = float;

	constexpr float COLOUR_MAX = 255.0F;
	constexpr uint32_t BYTE_MASK = 0xFF;
	constexpr uint8_t ONE_BYTE = 8;
	constexpr uint8_t TWO_BYTES = 16;

	constexpr math::vec3 rgb(uint32_t value) {
		uint8_t red = (value & BYTE_MASK);
		uint8_t green = ((value >> ONE_BYTE ) & BYTE_MASK);
		uint8_t blue = ((value >> TWO_BYTES ) & BYTE_MASK);
		return { (float)red / COLOUR_MAX, (float)green / COLOUR_MAX, (float)blue / COLOUR_MAX};
	}

	namespace colours {
		constexpr math::vec3 rgb_decimal(uint8_t red, uint8_t green, uint8_t blue) {
			return { (float)red / COLOUR_MAX, (float)green / COLOUR_MAX, (float)blue / COLOUR_MAX};
		}

		constexpr math::vec3 rgb(uint32_t value) {
			uint8_t red = ( (value >> TWO_BYTES) & BYTE_MASK);
			uint8_t green = ( (value >> ONE_BYTE) & BYTE_MASK);
			uint8_t blue = value & BYTE_MASK;
			return rgb_decimal(red, green, blue);
		}


		//
		// Extended (CSS3) Colours
		//

		// pinks
		constexpr auto MEDIUM_VIOLET_RED = rgb(0xC71585);
		constexpr auto DEEP_PINK = rgb(0xFF1493);
		constexpr auto PALE_VIOLET_RED = rgb(0xDB7093);
		constexpr auto HOT_PINK = rgb(0xFF69B4);
		constexpr auto LIGHT_PINK = rgb(0xFFB6C1);
		constexpr auto PINK = rgb(0xFFC0CB);

		// Reds
		constexpr auto DARK_RED = rgb(0x8B0000);
		constexpr auto RED = rgb(0xFF0000);
		constexpr auto FIRE_BRICK = rgb(0xB22222);
		constexpr auto CRIMSON = rgb(0xDC143C);
		constexpr auto INDIAN_RED = rgb(0xCD5C5C);
		constexpr auto LIGHT_CORAL = rgb(0xF08080);
		constexpr auto SALMON = rgb(0xFA8072);
		constexpr auto DARK_SALMON = rgb(0xE9967A);
		constexpr auto LIGHT_SALMON = rgb(0xFFA07A);

		// Oranges
		constexpr auto ORANGE_RED = rgb(0xFF4500);
		constexpr auto TOMATO = rgb(0xFF9347);
		constexpr auto DARK_ORANGE = rgb(0xFF8C00);
		constexpr auto CORAL = rgb(0xFF7F50);
		constexpr auto ORANGE = rgb(0xFFA500);

		// Yellows
		constexpr auto DARK_KHAKI = rgb(0xBDB76B);
		constexpr auto GOLD = rgb(0xFFD700);
		constexpr auto KHAKI = rgb(0xF0E68C);
		constexpr auto PEACH_PUFF = rgb(0xFFDAB9);
		constexpr auto YELLOW = rgb(0xFFFF00);
		constexpr auto PALE_GOLDENROD = rgb(0xEEE8AA);
		constexpr auto MOCCASIN = rgb(0xFFE4B5);
		constexpr auto PAPAYA_WHIP = rgb(0xFFEFD5);
		constexpr auto LIGHT_GOLDENROD_YELLOW = rgb(0xFAFAD2);
		constexpr auto LEMON_CHIFFON = rgb(0xFFFACD);
		constexpr auto LIGHT_YELLOW = rgb(0xFFFFE0);

		// Browns
		constexpr auto MAROON = rgb(0x800000);
		constexpr auto BROWN = rgb(0xA52A2A);
		constexpr auto SADDLE_BROWN = rgb(0x8B4513);
		constexpr auto SIENNA = rgb(0xA0522D);
		constexpr auto CHOCOLATE = rgb(0xD2691E);
		constexpr auto DARK_GOLDENROD = rgb(0xB8860B);
		constexpr auto PERU = rgb(0xCD853F);
		constexpr auto ROSY_BROWN = rgb(0xBC8F8F);
		constexpr auto GOLDENROD = rgb(0xDAA520);
		constexpr auto SANDY_BROWN = rgb(0xF4A460);
		constexpr auto TAN = rgb(0xD2B48C);
		constexpr auto BURLYWOOD = rgb(0xDEB887);
		constexpr auto WHEAT = rgb(0xF5DEB3);
		constexpr auto NAVAJO_WHITE = rgb(0xFFDEAD);
		constexpr auto BISQUE = rgb(0xFFEAC4);
		constexpr auto BLANCHED_ALMOND = rgb(0xFFEBCD);
		constexpr auto CORNSILK = rgb(0xFFF8DC);

		// Greens
		constexpr auto DARK_GREEN = rgb(0x006400);
		constexpr auto GREEN = rgb(0x008000);
		constexpr auto DARK_OLIVE_GREEN = rgb(0x556B2F);
		constexpr auto FOREST_GREEN = rgb(0x228B22);
		constexpr auto SEA_GREEN = rgb(0x2E8B57);
		constexpr auto OLIVE = rgb(0x808000);
		constexpr auto OLIVE_DRAB = rgb(0x6B8E23);
		constexpr auto MEDIUM_SEA_GREEN = rgb(0x3CB371);
		constexpr auto LIME_GREEN = rgb(0x32CD32);
		constexpr auto LIME = rgb(0x00FF00);
		constexpr auto SPRING_GREEN = rgb(0x00FF7F);
		constexpr auto MEDIUM_SPRING_GREEN = rgb(0x00FA9A);
		constexpr auto DARK_SEA_GREEN = rgb(0x8FBC8F);
		constexpr auto MEDIUM_AQUAMARINE = rgb(0x66CDAA);
		constexpr auto YELLOW_GREEN = rgb(0x9ACD32);
		constexpr auto LAWN_GREEN = rgb(0x7FFF00);
		constexpr auto CHARTREUSE = rgb(0x7FFF00);
		constexpr auto LIGHT_GREEN = rgb(0x90EE90);
		constexpr auto GREEN_YELLOW = rgb(0xADFF2F);
		constexpr auto PALE_GREEN = rgb(0x98FB98);

		// Cyans
		// TODO finish writing out the cyan list
		constexpr auto CYAN = rgb(0xFFFF);

		// Blues
		// TODO finish writing out the blue list
		constexpr auto MIDNIGHT_BLUE = rgb(0x191970);

		/// Purples, Violets, and Magentas
		// TODO finish writing out the purple list
		constexpr auto FUCHSIA = rgb(0xFF00FF);
		constexpr auto MAGENTA = FUCHSIA;

		// Whites
		//TODO finish writing out the white list
		constexpr auto WHITE = rgb(0xFFFFFF);

		// Grays and Blacks
		constexpr auto BLACK = rgb(0x000000);
		constexpr auto DARK_SLATE_GRAY = rgb(0x2F4F4F);
		constexpr auto DIM_GRAY = rgb(0x696969);
		constexpr auto SLATE_GRAY = rgb(0x708090);
		constexpr auto GRAY = rgb(0x808080);
		constexpr auto LIGHT_SLATE_GRAY = rgb(0x778899);
		constexpr auto DARK_GRAY = rgb(0xA9A9A9);
		constexpr auto SILVER = rgb(0xC0C0C0);
		constexpr auto LIGHT_GRAY = rgb(0xD3D3D3);
		constexpr auto GAINSBORO = rgb(0xDCDCDC);
	}



	enum class PathType {
		MOVE,
		PATH,
		BAZIER2,
		BAZIER3,
		COLOUR,
		CLOSE
	};

	struct Path2D {
		inline explicit Path2D(math::vec2 start) : m_points(), m_types() {
			moveTo(start);
		}

		inline void moveTo(math::vec2 pos) {
			m_points.push_back(pos);
			m_types.push_back(PathType::MOVE);
		}

		inline void pathTo(math::vec2 pos) {
			m_points.push_back(pos);
			m_types.push_back(PathType::PATH);
		}

		void bezierTo(math::vec2 cp1, math::vec2 pos);
		void bezierTo(math::vec2 cp1, math::vec2 cp2, math::vec2 pos);

		void arc(math::vec2 center, float radius, RadianAngle startAngle, RadianAngle endAngle, bool ccw);

		void colour(math::vec3 colour) {
			m_colours.push_back(colour);
			m_types.push_back(PathType::COLOUR);
		}

		void close() {
			pathTo(m_points[0]);
			m_types.push_back(PathType::CLOSE);
		}

		std::vector<math::vec2> m_points;
		std::vector<PathType> m_types;
		std::vector<math::vec3> m_colours;
	};

	inline Path2D make_rect(math::vec2 center, math::vec2 radius, math::vec3 colour = {1.0f, 1.0f, 1.0f}) {
		fggl::gfx::Path2D tileGfx(center);
		tileGfx.colour(colour);
		tileGfx.moveTo(center - radius);
		tileGfx.pathTo({center.x + radius.x, center.y - radius.y});
		tileGfx.pathTo(center + radius);
		tileGfx.pathTo({center.x - radius.x, center.y + radius.y});
		tileGfx.close();
		return tileGfx;
	}

	struct ShapeOpts {
		float angleOffset = 0.0F;
		bool sinFirst = true;
	};

	inline Path2D make_shape(math::vec2 center, float radius, int sides, math::vec3 colour = colours::WHITE, ShapeOpts opts = {}) {
		float angle = (math::PI * 2.0F) / sides;

		fggl::gfx::Path2D tileGfx(center);
		tileGfx.colour(colour);

		for (int i=0; i < sides; ++i) {
			float xPos = (float)(sinf(i * angle + opts.angleOffset) * radius) + center.x;
			float yPos = (float)(cosf(i * angle + opts.angleOffset) * radius) + center.y;
			if (!opts.sinFirst) {
				std::swap(xPos, yPos);
			}
			if ( i == 0 ) {
				tileGfx.moveTo( {xPos, yPos} );
			} else {
				tileGfx.pathTo({xPos, yPos});
			}
		}
		tileGfx.close();
		return tileGfx;
	}

	inline Path2D make_arc(math::vec2 center, float radius, float start, float end, math::vec3 colour = colours::WHITE, int slices = 25) {
		const float angle = (M_PI * 2.0) / slices;
		fggl::gfx::Path2D tileGfx(center);

		tileGfx.colour(colour);
		tileGfx.moveTo(center);
		for (float totalAngle = start; totalAngle < end; totalAngle += angle) { // NOLINT(cert-flp30-c)
			float xPos = (float)(sinf(totalAngle) * radius) + center.x;
			float yPos = (float)(cosf(totalAngle) * radius) + center.y;
			tileGfx.pathTo({xPos, yPos});
			if ( (totalAngle + angle) > end) {
				break;
			}
		}

		{
			float xPos = (float) (sinf(end) * radius) + center.x;
			float yPos = (float) (cosf(end) * radius) + center.y;
			tileGfx.pathTo( {xPos, yPos} );
		}

		tileGfx.close();
		return tileGfx;
	}

	enum class PaintType {
		FILL,
		STROKE
	};

	struct PaintCmd {
		PaintType type;
		Path2D path;
	};

	struct TextCmd {
		const std::string text;
		const math::vec2 pos;
		const math::vec3 colour;
	};

	class Paint {
		public:
			Paint() = default;
			Paint(Paint &paint) = delete;

			inline void fill(Path2D &path) {
				m_cmds.push_back({PaintType::FILL, path});
			}

			void stroke(Path2D &path) {
				m_cmds.push_back({PaintType::STROKE, path});
			}

			void text(const std::string &text, const math::vec2 &pos, const math::vec3f colour = fggl::gfx::colours::BLACK) {
				m_text.push_back({text, pos, colour});
			}

			const std::vector<PaintCmd> &cmds() const {
				return m_cmds;
			}

			const std::vector<TextCmd> &textCmds() const {
				return m_text;
			}

		private:
			std::vector<PaintCmd> m_cmds;
			std::vector<TextCmd> m_text;
	};

}

#endif
